: # Find Perl
eval 'exec perl5 -w -S $0 "$@"' 
if 0;

# This is a simple demo script, and adds zero bits of 
# information beyond the README and other files.  Since
# it serves no purpose but to give people a way to test
# in one command, even w/o reading anything else, this
# script has a few additional dependencies.
#
#
# - This file may be rife with Unixisms.
# - Gnuplot should be in the user's path.
# - Ghostview is required for display

# This is more of a generic post-install script than
# anything you'd want to take apart and include in your
# own code.


# Just making the include explicit.
use lib '.';

use strict;
use English;
use Symbol;
use FileHandle;

require DiskFracDim;
require DiskWrapper;
use     mktemp;

# Certain phases (namely, the FracDim computation itself)
# may take a considerable amount of time.  Don't buffer 
# STDOUT, because otherwise the delay may be considerable.
#
autoflush STDOUT 1;



# Phase 1 -- Permissions.
#
# Since we need to modify a couple o' 'use lib' statements,
# we want to make sure that two of the files are writeable.
# If they aren't, toggle those bits on.  
#
# This phase knows nothing of ACLs, just traditional Unix
# fast-filesystem bits.
#
# We also need to set the executable bit on all included .pl files.
{
    # 0100 = user can execute
    # 0200 = user can write
    # 
    # We fix ourselves in case we weren't shipped as writeable
    # but a user managed to invoke it via, say, ". RUNME.pl".
    # Just for future convenience.
    #

    
    print "*** Checking file permissions ***\n";
 
    if (!(-w ".")) {
	die "Fatal error -- cannot write to current directory.";
    }

    my @fix_files = ( 
		     +[ "DiskFracDim.pm",  0200 ],
		     +[ "DiskFracTest.pl", 0300 ],
		     +[ "FixPlace.pl",     0100 ],
		     +[ "RUNME.pl",        0100 ],
		     +[ "LineGen.pl",      0100 ],
		     +[ "TriGen.pl",       0100 ]
		     );

    my $listref = undef;
    foreach $listref (@fix_files) {
	my $file = $listref->[0];
	my $goal = $listref->[1];

	print "Checking permissions on $file ($goal)... ";

	if (!(-e $file)) {
	    die "Fatal error -- expected file $file does not exist.";
	}

	# Reuse stat info from the last test
	my ($dev,   $ino,   $mode,  $nlink,
	    $uid,   $gid,   $rdev,  $size,
	    $atime, $mtime, $ctime, $blksize,
	    $blocks) = stat($file);

	if ($uid ne $EUID) {
	    die "Fatal error -- '$file' owned by uid $uid.  EUID = $EUID.";
	}

	if (($mode & $goal) == $goal) {
	    print "ok\n";
	    next;
	} else {
	    print "off... ";
	    
	    $mode |= $goal;
	    if (!(chmod $mode, $file)) {
		die "Fatal error:  chmod() failed on $file:  $!";
	    }

	    print "(fixed)\n";
	}
    }

}



# Phase 2 -- Fixing library calls.
#
# Now, we invoke the FixPlace script.
#
# 
{
    print "*** Running FixPlace.pl ***\n";

    # This really should NOT fail, if we made it past the
    # previous phase...
    if (system("FixPlace.pl")) {
	die "Error running FixPlace.pl:  $!";
    }
}



# Phase 3 -- Generating test data.
#
# Generating a Sierpinski data set, of 5,000 points 
# only (for speed reasons.  Given that this script
# might be the first thing run, it behooves us to
# not use a terribly slow case.
{
    print "*** Generating Sierpinski.5000 ***\n";

    if (system("./TriGen.pl 5000 > Sierpinski.5000")) {
	die "Error running TriGen.pl:  $!";
    }
}


# Phase 4 -- Invoke FracDim code.  
#
# We grab the log-log results directly from the structure.
# Outside scripts might simply want to grab stdout and
# parse it (see the format used in 'print_loglog()'),
# or alternately add an explicit method for taking the
# pairs from the DiskFracDim object.


# This will be a temporary file to feed into gnuplot as
# data.
my $pair_fn = undef;
my $fd      = undef;
my $y_int   = undef;

{

    print "*** Invoking FracDim *** \n";
    print "\n";
    print "WARNING -- This phase may take a considerable\n";
    print "amount of time.  On a 400MHz Pentium II, this\n";
    print "took approximately 44 seconds.\n";
    print "\n";
    print "Due to optimizations that do not apply to the\n";
    print "Sierpinski set, you may find that the speed\n";
    print "varies considerably depending on the data.\n";
    print "\n";

    my $sier_fh = gensym();
    my $wrapper = undef;
    my $fracobj = undef;
    my %params  = ();

    # We need a filehandle to pass to the wrapper.

    open($sier_fh, "Sierpinski.5000") || 
	die "Strange, fatal error opening Sierpinski.5000:  $!";

    $wrapper = DiskWrapper::new DiskWrapper($sier_fh);

    $params{'data_o'}   = $wrapper;
    $params{'verbose'}  = 1;
    $fracobj = DiskFracDim::new DiskFracDim(%params);
    
    my $corr = undef;
    ($fd, $y_int, $corr) = $fracobj->fracdim();
    
    print "Fractal dimension reported:  $fd (theoretical ~ 1.58)\n";
    print "(line-fitting) Correlation:  $corr\n";

    close($sier_fh) || 
	die "Unexpected fatal error closing Sierpinski.5000:  $!";


    # Allow for some precision issues due to architectural 
    # differences, but .001 should probably be a safe level of
    # accuracy to demand.
    if ((int($fd * 1000)/1000) != 1.627) {
	die "Fatal error -- expected the fd value to be ~1.627.";
    }
 
    my @log_r = @{$fracobj->{'data'}->{'log_r'}};
    my @log_s = @{$fracobj->{'data'}->{'log_s'}};

    # Store this for Gnuplot to use.
    my $pair_fh = undef;
    my $ct = scalar @log_r;
    my $i  = undef;

    ($ct == (scalar @log_s)) || 
	die "Severe, unexpected logical error in DiskFracDim!";
    ($pair_fh, $pair_fn) = mktemp();

    # It's log data, but we'll use gnuplot's scale to make it
    # clear, so we undo the logs.
    for ($i=0; $i < $ct; $i++) {
	print $pair_fh (2 ** ($log_r[$i])), " ", (2 ** ($log_s[$i])), "\n";
    }

    close($pair_fh) || die "Failed to write pair-count tempfile:  $!";
}



# Phase 5 -- Gnuplot.
#
# We use the data obtained from the previous section to plot.
# Note that the coordinates have already been trimmed to remove
# 'flat' regions at either extreme.

{
    my $gnuplot_fh = gensym();

    print "*** Using GNUPLOT for log-log graph ***\n";

    if (!open($gnuplot_fh, "|gnuplot")) {
	die "Fatal error -- gnuplot not in path?  $!";
    }

    unlink("sierpinski.eps");

    print $gnuplot_fh "set terminal postscript eps\n";    
    print $gnuplot_fh "set output \"sierpinski.eps\"\n";
    print $gnuplot_fh "set title \"Pair-count plot for Sierpinski.5000\"\n";
    print $gnuplot_fh "set xlabel \"Radius (logarithmic)\"\n";
    print $gnuplot_fh "set ylabel \"Pairs (logarithmic)\"\n";
    print $gnuplot_fh "set logscale xy 2\n";
    print $gnuplot_fh "set key under\n";
    print $gnuplot_fh "plot \"$pair_fn\" title \"Sierpinski\" with linespoints\n";
    print $gnuplot_fh "replot (2**(($fd*log(x)/log(2))+$y_int)) title \"Linear approximation\"\n";
    print $gnuplot_fh "set terminal postscript eps\n";    
    print $gnuplot_fh "set output \"sierpinski.eps\"\n";
    print $gnuplot_fh "replot\n";
    close($gnuplot_fh) || die "Unexpected error -- failed to end gnuplot:  $!";
}


# Phase 6 -- Display

{
    print "*** Displaying resulting sierpinski.eps ***\n";
    
    system("gv sierpinski.eps") && die "Unexpected ghostview return value";
}



# Phase 7 -- Generating test data.
#
# Generating a self-similar line (with gaps).
#
# Warning -- it's 50-dimensional and thus this takes ~4MB space or so.
#
{
    print "*** Generating Line.5000 ***\n";

    if (system("./LineGen.pl 5000 > Line.5000")) {
	die "Error running LineGen.pl:  $!";
    }
}



# Phase 8 -- Invoke FracDim code with q=1 (special case).
#
# Very similar to Phase 4, but it tests code that only applies
# to q=1, and it uses a different data set...


# This will be a temporary file to feed into gnuplot as
# data.

{

    print "*** Invoking FracDim for Q=1 *** \n";
    print "\n";
    print "WARNING -- This phase may also take a considerable\n";
    print "amount of time.\n";
    print "\n";
    print "This test invokes code that only applies for\n";
    print "the special case of the exponent being 1.  The\n";
    print "results should in this case be slightly \n";
    print "different.\n";
    print "\n";

    my $sier_fh = gensym();
    my $wrapper = undef;
    my $fracobj = undef;
    my %params  = ();

    # We need a filehandle to pass to the wrapper.

    open($sier_fh, "Line.5000") || 
	die "Strange, fatal error opening Line.5000:  $!";

    $wrapper = DiskWrapper::new DiskWrapper($sier_fh);

    $params{'data_o'}   = $wrapper;
    $params{'q'}        = 1;
    $params{'verbose'}  = 1;
    $fracobj = DiskFracDim::new DiskFracDim(%params);
    
    my $corr = undef;
    ($fd, $y_int, $corr) = $fracobj->fracdim();
    
    print "Fractal dimension reported:  $fd (theoretical ~ 0.5)\n";
    print "(line-fitting) Correlation:  $corr\n";

    close($sier_fh) || 
	die "Unexpected fatal error closing Line.5000:  $!";


    # Allow for some precision issues due to architectural 
    # differences, but .001 should probably be a safe level of
    # accuracy to demand.
    if ((int($fd * 1000)/1000) != 0.498) {
	die "Fatal error -- expected the fd value to be ~0.498.";
    }
 
    my @log_r = @{$fracobj->{'data'}->{'log_r'}};
    my @log_s = @{$fracobj->{'data'}->{'log_s'}};

    # Store this for Gnuplot to use.
    my $pair_fh = undef;
    my $ct = scalar @log_r;
    my $i  = undef;

    ($ct == (scalar @log_s)) || 
	die "Severe, unexpected logical error in DiskFracDim!";
    ($pair_fh, $pair_fn) = mktemp();

    # It's log data, but we'll use gnuplot's scale to make it
    # clear, so we undo the logs.
    #
    # Well, in this case, the latter wasn't actually passed
    # through a logarithm.  Things just get a bit odd for q=1.
    for ($i=0; $i < $ct; $i++) {
	print $pair_fh (2**$log_r[$i]), " ", $log_s[$i], "\n";
    }

    close($pair_fh) || die "Failed to write pair-count tempfile:  $!";
}



# Phase 9 -- Gnuplot.
#
# We use the data obtained from the previous section to plot.
# Note that the coordinates have already been trimmed to remove
# 'flat' regions at either extreme.

{
    my $gnuplot_fh = gensym();

    print "*** Using GNUPLOT for log-log graph ***\n";

    if (!open($gnuplot_fh, "|gnuplot")) {
	die "Fatal error -- gnuplot not in path?  $!";
    }

    unlink("line.eps");

    print $gnuplot_fh "set terminal postscript eps\n";    
    print $gnuplot_fh "set output \"line.eps\"\n";
    print $gnuplot_fh "set title \"Pair-count plot for Line.5000\"\n";
    print $gnuplot_fh "set xlabel \"Radius (logarithmic)\"\n";
    print $gnuplot_fh "set ylabel \"Negative entropy\"\n";
    print $gnuplot_fh "set logscale x 2\n";
    print $gnuplot_fh "set key under\n";
    print $gnuplot_fh "plot \"$pair_fn\" title \"Line\" with linespoints\n";
    print $gnuplot_fh "replot (($fd*log(x)/log(2))+$y_int) title \"Linear approximation\"\n";
    print $gnuplot_fh "set terminal postscript eps\n";    
    print $gnuplot_fh "set output \"line.eps\"\n";
    print $gnuplot_fh "replot\n";
    close($gnuplot_fh) || die "Unexpected error -- failed to end gnuplot:  $!";
}


# Phase 10 -- Display

{
    print "*** Displaying resulting line.eps ***\n";
    
    system("gv line.eps") && die "Unexpected ghostview return value";
}


# Phase 11 -- Final comments

{
    print "\n";
    print "Your installation appears to be in functioning\n";
    print "properly.  See the README for more information on\n";
    print "using this package.\n";
}
